import {
  serializePlaybill,
  ValidatedPlaybillSchema,
} from "@proscenium/playbill";
import chalk from "chalk";
import { watch } from "node:fs/promises";
import { loadFromBinding, resolveBindingPath } from "./filetypes/binding";
import { usage } from "./usage";
import { CLIError, MuseError } from "./errors";
import { renderAsHTML } from "./renderers/html";
import { parseCLIArguments, type CLIArguments } from "./args";
import { fullDirname } from "./util/files";

enum ExitCode {
  Success = 0,
  Error = 1,
}

async function processBinding({ inputFilePath, options }: CLIArguments) {
  // -- BINDING FILE -- //
  const bindingPath = await resolveBindingPath(inputFilePath);
  const binding = await loadFromBinding(bindingPath);

  // -- VALDATE PLAYBILL -- //
  const validatedPlaybill = ValidatedPlaybillSchema.safeParse(binding.playbill);

  if (!validatedPlaybill.success) {
    console.error("Error validating playbill");
    console.error(validatedPlaybill.error.errors);
    return ExitCode.Error;
  }

  // -- EXIT EARLY IF JUST CHECKING VALIDATION --//
  if (options.check) {
    console.log(chalk.green("Playbill validated successfully"));
    return ExitCode.Success;
  }

  // -- SERIALIZE USING RENDERER --//
  let serializedPlaybill = "";

  switch (options.renderer) {
    case "json":
      serializedPlaybill = serializePlaybill(validatedPlaybill.data);
      break;
    case "html":
      serializedPlaybill = await renderAsHTML(validatedPlaybill.data, binding);
      break;
    default:
      throw new CLIError(`Unknown renderer: ${options.renderer}`);
  }

  // -- WRITE TO DISK OR STDOUT --//
  if (options.write) {
    await Bun.write(options.outfile, serializedPlaybill);

    return ExitCode.Success;
  }

  console.log(serializedPlaybill);
  return ExitCode.Success;
}

async function main(): Promise<number> {
  const cliArguments = parseCLIArguments(Bun.argv.slice(2));
  const { options } = cliArguments;

  // -- HELP TEXT -- //
  if (options.help) {
    console.log(usage);
    return ExitCode.Success;
  }

  // -- ARG VALIDATION -- //
  if (options.check && options.write) {
    throw new CLIError("Cannot use --check and --write together");
  }

  console.log(`Processing ${cliArguments.inputFilePath}`);
  let lastProcessResult = await processBinding(cliArguments);

  if (options.watch) {
    const watchDir = fullDirname(cliArguments.inputFilePath);
    console.log(`Watching ${watchDir} for changes`);
    const watcher = watch(watchDir, {
      recursive: true,
    });

    for await (const event of watcher) {
      console.log(`Detected ${event.eventType} on ${event.filename}`);
      lastProcessResult = await processBinding(cliArguments);

      if (lastProcessResult === ExitCode.Error) {
        console.error(`Error processing ${event.filename}`);
      } else {
        console.log("Reprocessed changes");
      }
    }
    return ExitCode.Success;
  }

  return ExitCode.Success;
}

try {
  const exitCode = await main();
  process.exit(exitCode);
} catch (error) {
  if (error instanceof MuseError) {
    console.error(chalk.red(error.fmt()));
  } else {
    console.error("An unexpected error occurred");
    console.error(error);
  }

  process.exit(1);
}
